/*
 * Copyright (c) 2005-2013, BearWare.dk
 * 
 * Contact Information:
 *
 * Bjoern D. Rasmussen
 * Bedelundvej 79
 * DK-9830 Taars
 * Denmark
 * Email: contact@bearware.dk
 * Phone: +45 20 20 54 59
 * Web: http://www.bearware.dk
 *
 * This source code is part of the TeamTalk 4 SDK owned by
 * BearWare.dk. All copyright statements may not be removed 
 * or altered from any source distribution. If you use this
 * software in a product, an acknowledgment in the product 
 * documentation is required.
 *
 */

#include "uservideowidget.h"
#include "common.h"
#include "preferencesdlg.h"

#include <QPaintEngine>
#include <QDebug>
#include <QTimerEvent>

extern TTInstance* ttInst;
extern QSettings* ttSettings;

#define USE_VIDEO_SHAREDMEM_COPY 1
#define USE_VIDEO_SHAREDMEM_ZEROCOPY 0 //TODO: UserVideoWidget can be active 
                                       //for the same user twice due to detached 
                                       //UserVideoDlg-class which isn't deleted 
                                       //until the message loop has finished 
                                       //(Qt::WA_DeleteOnClose). The destructor
                                       //then deletes BMP-data which is active
                                       //in another QImage => crash

UserVideoWidget::UserVideoWidget(QWidget * parent, 
                                 int userid/* = -1 */,
                                 int stream_id/* = 0 */,
                                 int border_width/* = 0 */)
  : UserImageWidget(parent, userid, true, border_width)
  , m_streamid(stream_id)
  , m_novideo_count(0)
{
    m_vidtext = ttSettings->value(SETTINGS_DISPLAY_VIDEOTEXT,
                                  SETTINGS_DISPLAY_VIDEOTEXT_DEFAULT).toUInt();
        
    m_vidtext_w_pct = ttSettings->value(SETTINGS_DISPLAY_VIDEOTEXT_WIDTH,
                                        SETTINGS_DISPLAY_VIDEOTEXT_WIDTH_DEFAULT).toInt();
    m_vidtext_h_pct = ttSettings->value(SETTINGS_DISPLAY_VIDEOTEXT_HEIGHT,
                                        SETTINGS_DISPLAY_VIDEOTEXT_HEIGHT_DEFAULT).toInt();

    m_vidtext_bg = ttSettings->value(SETTINGS_DISPLAY_VIDEOTEXT_BGCOLOR,
                                     SETTINGS_DISPLAY_VIDEOTEXT_BGCOLOR_DEFAULT)
                                     .value<QColor>();
    m_vidtext_fg = ttSettings->value(SETTINGS_DISPLAY_VIDEOTEXT_FGCOLOR,
                                     SETTINGS_DISPLAY_VIDEOTEXT_FGCOLOR_DEFAULT)
                                     .value<QColor>();
    startTimer(1000);
}

UserVideoWidget::~UserVideoWidget()
{
#if USE_VIDEO_SHAREDMEM_ZEROCOPY
    TT_ReleaseUserVideoFrame(ttInst, m_userid);
#endif
}

QSize UserVideoWidget::imageSize() const
{
#ifdef USE_TT_PAINT
    CaptureFormat format;
    if(!TT_GetUserVideoFrame(ttInst, m_userid, NULL, 0, &format))
        return QSize();
    return QSize(format.nWidth, format.nHeight);
#else
    return m_image.size();
#endif
}

#ifdef USE_TT_PAINT
void UserVideoWidget::runTTPaint(QPainter& painter)
{
    HDC hdc = painter.paintEngine()->getDC();
    QPoint p0 = this->mapTo(nativeParentWidget(), QPoint(0,0));
    //drain frames (we don't want any queued frames)
    while(TT_PaintVideoFrame(ttInst, m_userid, hdc, p0.x(), p0.y(), width(), height()));
    painter.paintEngine()->releaseDC(hdc);
}
#endif

void UserVideoWidget::slotNewVideoFrame(int userid, int frame_count)
{
    if(m_userid != userid)
        return;

    m_novideo_count = 0;

#ifndef USE_TT_PAINT
    bool modified = false;
    //drain frames (we don't want any queued frames)
    while(frame_count--)
    {
#if USE_VIDEO_SHAREDMEM_COPY
        VideoFrame vidframe;
        if(TT_AcquireUserVideoFrame(ttInst, m_userid, &vidframe))
        {
            m_image = QImage(vidframe.nWidth, vidframe.nHeight,
                QImage::Format_RGB32);
            Q_ASSERT(vidframe.nFrameBufferSize == m_image.byteCount());
            memcpy(m_image.bits(), vidframe.frameBuffer, 
                vidframe.nFrameBufferSize);
            m_streamid = vidframe.nStreamID;

            TT_ReleaseUserVideoFrame(ttInst, m_userid);
        }
#elif USE_VIDEO_SHAREDMEM_ZEROCOPY
#pragma error "Zero copy doesn't work yet"
        //Zero-Copy (shared mem).
        //Use memory allocated by TeamTalk DLL as raw input to QImage
        TT_ReleaseUserVideoFrame(ttInst, m_userid);

        VideoFrame vidframe;
        if(TT_AcquireUserVideoFrame(ttInst, m_userid, &vidframe) &&
            (m_streamid == vidframe.nStreamID || vidframe.bKeyFrame))
        {
            m_image = QImage(static_cast<uchar*>(vidframe.frameBuffer),
                             vidframe.nWidth, vidframe.nHeight,
                             QImage::Format_RGB32);

            m_streamid = vidframe.nStreamID;
        }
#else
        //Copy mem.
        //Preallocate memory and make TeamTalk DLL fill the allocated
        //memory with image data.
        CaptureFormat format;
        if(!TT_GetUserVideoFrame(ttInst, userid, NULL, 0, &format))
            break;

        Q_ASSERT(format.picFourCC == FOURCC_RGB32);

        if(m_image.width() != format.nWidth || m_image.height() != format.nHeight)
            m_image = QImage(format.nWidth, format.nHeight, QImage::Format_RGB32);

        if(!TT_GetUserVideoFrame(ttInst, userid, m_image.bits(), m_image.numBytes(), &format))
            break;
#endif

        modified = true;
    }

    if(modified)
        update();
#else
    Q_UNUSED(frame_count);
    update();
#endif
}

void UserVideoWidget::timerEvent(QTimerEvent *e)
{
    if(m_novideo_count++ >= VIDEOSESSION_TIMEOUT)
    {
        killTimer(e->timerId());
        emit(userVideoEnded(m_userid));
    }
}

void UserVideoWidget::completePaint(const QRect& r, QPainter& painter)
{
    if(m_vidtext != VIDEOTEXT_NONE)
    {
        User user;
        if(TT_GetUser(ttInst, m_userid, &user))
        {
            QString text = QString("User #%1").arg(m_userid);
            switch(m_vidtext & VIDTEXT_SHOW_MASK)
            {
            case VIDTEXT_SHOW_NICKNAME :
                text = _Q(user.szNickname);
                break;
            case VIDTEXT_SHOW_USERNAME :
                text = _Q(user.szUsername);
                break;
            case VIDTEXT_SHOW_STATUSTEXT :
                text = _Q(user.szStatusMsg);
                break;
            }
            setVideoTextBox(r, m_vidtext_bg, m_vidtext_fg, text, m_vidtext, 
                            m_vidtext_w_pct, m_vidtext_h_pct, painter);
        }
    }
}
